/*******************************************************************************

Copyright (c) 1997-2004, Perforce Software, Inc.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTR
IBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE SOFTWARE, INC. BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

/*******************************************************************************
 * Name		: clientuserruby.cc
 *
 * Author	: Tony Smith <tony@perforce.com> or <tony@smee.org>
 *
 * Description	: Ruby bindings for the Perforce API. User interface class
 * 		  for getting Perforce results into Ruby.
 *
 ******************************************************************************/
#include <ctype.h>
#include <ruby.h>
#include <clientapi.h>
#include <spec.h>
#include <diff.h>
#include "gc_hack.h"
#include "p4result.h"
#include "clientuserruby.h"
#include "specmgr.h"


/*******************************************************************************
 * ClientUserRuby - the user interface part. Gets responses from the Perforce
 * server, and converts the data to Ruby form for returning to the caller.
 ******************************************************************************/

ClientUserRuby::ClientUserRuby()
{ 
    debug = 0;
    input = Qnil;
}


void
ClientUserRuby::Reset()
{
    results.Reset();
    lastSpecDef.Clear();

    // Leave input alone.
}

void	
ClientUserRuby::Finished()
{
    // Reset input coz we should be done with it now. Keeping hold of
    // it just prevents GC from sweeping it if possible
    if ( debug && input != Qnil )
	fprintf( stderr, "[P4] Cleaning up saved input\n" );

    input = Qnil;
}

void
ClientUserRuby::HandleError( Error *e )
{
    if ( debug )
	fprintf( stderr, "[P4] HandleError()\n" );

    results.AddError( e );
}

void
ClientUserRuby::OutputText( const_char *data, int length )
{
    if ( debug )
	fprintf( stderr, "[P4] OutputText()\n" );

    results.AddOutput( data );
}

void
ClientUserRuby::OutputInfo( char level, const_char *data )
{
    if ( debug )
	fprintf( stderr, "[P4] OutputInfo()\n" );

    results.AddOutput( data );
}

void
ClientUserRuby::OutputBinary( const_char *data, int length )
{
    if ( debug )
	fprintf( stderr, "[P4] OutputBinary() %d bytes\n", length );

    //
    // Binary is just stored in a string. Since the char * version of
    // P4Result::AddOutput() assumes it can strlen() to find the length,
    // we'll make the String object here.
    //
    results.AddOutput( rb_str_new( data, length) );
}

void
ClientUserRuby::OutputStat( StrDict *values )
{
    SpecMgr	m;
    StrPtr	*spec, *data;

    // If both specdef and data are set, then we need to parse the form
    // and return the results. If not, then we just convert it as is.

    spec = values->GetVar( "specdef" );
    data = values->GetVar( "data" );

    if ( spec && data )
    {
	if ( debug )
	    fprintf( stderr, "[P4] OutputStat() - parsing form\n" );

	//
	// Save the spec definition for later retrieval by P4ClientApi
	//
	lastSpecDef = spec->Text();

	// Parse up the form. Use the ParseNoValid() interface to prevent
	// errors caused by the use of invalid defaults for select items in
	// jobspecs.
	SpecDataTable	specData;
	Spec		s( spec->Text(), "" );
	Error		e;

	s.ParseNoValid( data->Text(), &specData, &e );
	if ( e.Test() )
	{
	    HandleError( &e );
	    return;
	}

	results.AddOutput( m.DictToSpec( specData.Dict(), spec ) );
    }
    else
    {
	if ( debug )
	    fprintf(stderr ,"[P4] OutputStat() - converting StrDict to hash\n");

	results.AddOutput( m.DictToHash( values ) );
    }
}


/*
 * Diff support for Ruby API. Since the Diff class only writes its output
 * to files, we run the requested diff putting the output into a temporary
 * file. Then we read the file in and add its contents line by line to the 
 * results.
 */

void
ClientUserRuby::Diff( FileSys *f1, FileSys *f2, int doPage, 
				char *diffFlags, Error *e )
{

    if ( debug )
	fprintf( stderr, "[P4] Diff() - comparing files\n" );

    //
    // Duck binary files. Much the same as ClientUser::Diff, we just
    // put the output into Ruby space rather than stdout.
    //
    if( !f1->IsTextual() || !f2->IsTextual() )
    {
	if ( f1->Compare( f2, e ) )
	    results.AddOutput( "(... files differ ...)" );
	return;
    }

    // Time to diff the two text files. Need to ensure that the
    // files are in binary mode, so we have to create new FileSys
    // objects to do this.

    FileSys *f1_bin = FileSys::Create( FST_BINARY );
    FileSys *f2_bin = FileSys::Create( FST_BINARY );
    FileSys *t = FileSys::CreateGlobalTemp( f1->GetType() );

    f1_bin->Set( f1->Name() );
    f2_bin->Set( f2->Name() );

    {
	//
	// In its own block to make sure that the diff object is deleted
	// before we delete the FileSys objects.
	//
#ifndef OS_NEXT
	::
#endif
	Diff d;

	d.SetInput( f1_bin, f2_bin, diffFlags, e );
	if ( ! e->Test() ) d.SetOutput( t->Name(), e );
	if ( ! e->Test() ) d.DiffWithFlags( diffFlags );
	d.CloseOutput( e );

	// OK, now we have the diff output, read it in and add it to 
	// the output.
	if ( ! e->Test() ) t->Open( FOM_READ, e );
	if ( ! e->Test() ) 
	{
	    StrBuf 	b;
	    while( t->ReadLine( &b, e ) )
		results.AddOutput( b.Text() );
	}
    }

    delete t;
    delete f1_bin;
    delete f2_bin;

    if ( e->Test() ) HandleError( e );
}


/*
 * convert input from the user into a form digestible to Perforce. This
 * involves either (a) converting any supplied hash to a Perforce form, or
 * (b) running to_s on whatever we were given. 
 */

void
ClientUserRuby::InputData( StrBuf *strbuf, Error *e )
{
    if ( debug )
	fprintf( stderr, "[P4] InputData(). Using supplied input\n" );

    VALUE  	inval = input;

    if( Qtrue == rb_obj_is_kind_of( input, rb_cArray ) )
    {
	inval = rb_ary_shift( input );
    }

    if( Qnil == inval )
    {
	rb_warn( "[P4] Expected user input, found none. " 
		 "Missing call to P4#input()?" );
	return;
    }

    if ( Qtrue == rb_obj_is_kind_of( inval, rb_cHash ) )
    {
	SpecMgr 	m;
	StrPtr * 	specDef = varList->GetVar( "specdef" );

	m.HashToText( inval, strbuf, specDef );
	return;
    }

    // Convert whatever's left into a string
    ID	to_s 	= rb_intern( "to_s" );
    VALUE	str 	= rb_funcall( inval, to_s, 0 );
    strbuf->Set( STR2CSTR( str ) );
}

/*
 * In a script we don't really want the user to see a prompt, so we
 * (ab)use the SetInput() function to allow the caller to supply the
 * answer before the question is asked.
 */
void
ClientUserRuby::Prompt( const StrPtr &msg, StrBuf &rsp, int noEcho, Error *e )
{
    if ( debug )
	fprintf( stderr, "[P4] Prompt(): %s\n", msg.Text() );

    InputData( &rsp, e );
}


/*
 * Accept input from Ruby and convert to a StrBuf for Perforce
 * purposes.  We just save what we're given here because we may not 
 * have the specdef available to parse it with at this time.
 */

VALUE
ClientUserRuby::SetInput( VALUE i )
{
    if ( debug )
	fprintf( stderr, "[P4] SetInput()\n" );

    input = i;
    return Qtrue;
}


//
// GC support
//
void
ClientUserRuby::GCMark()
{
    if ( debug >= 2 )
    	fprintf( stderr, 
		"[P4] Marking results and errors for garbage collection\n" );

    if ( input != Qnil ) 
	rb_gc_mark( input );

    results.GCMark();
}
